/**
 *
 * InstanceDiskArrayList.java -- used in SEuS structure discovery system
 * Copyright (C) 2002 Shayan Ghazizadeh
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


import java.io.*;
import java.lang.*;
import java.util.*;

/** 
 * This class implements an array of instances of
 * changing size. The array is partly in main memory
 * and partly saved on disk.
 *
 * @author Shayan Ghazizadeh <shayan@cs.umd.edu>
 */
class InstanceDiskArrayList implements Serializable{

    private ArrayList pages;
    private int maxPageInMem;
    private int pagesInMem;
    private int pageSize;
    private int nextIndex;
    private int nextPage;
    private boolean getMode;
    private boolean sleep;

    private static int totalAccess=0;
    private static int pageFaults=0;
    private int fileCount;
    /**
     * Constructor.
     *
     * @param m  Maximum number of pages to have in main memory
     *           at any time.
     * @param ps size of each page.
     */
    public InstanceDiskArrayList(int m, int ps) {
	pages = new ArrayList();
	maxPageInMem = m;
	pageSize = ps;
	nextIndex = 0;
	nextPage = 0;
	pagesInMem = 0;

	getMode = false;
	fileCount = 0;
    }

    public InstanceDiskArrayList(boolean empty) {
	nextIndex = 0;
	nextPage = 0;
	getMode = true;
	fileCount = 0;
	pagesInMem = 0;
	pageSize = 0;
	maxPageInMem = 0;
	pages = null;
    }

    /**
     * This method is used to change the maximum number of pages in Mem.
     * It can only be used to increase the number.
     *
     * @param m New maximum number of pages in memory.
     */
    public void changeMPM(int m) {
	if (m <= maxPageInMem) {
	    System.err.println("InstanceDiskArrayList.changeMPM(): Invalid change request. Ignored.");
	}
	else
	    maxPageInMem = m;
    }


    /**
     * This method returns the current number of objects in our array.
     */
    public int size() {
	return nextIndex;
    }


    public void go2Sleep() throws SleepWhileInAddModeException {
	//if (!getMode) 
	//  throw new SleepWhileInAddModeException();
	sleep = true;
	if (nextPage != 0) {
	    Page x = (Page)pages.get(nextPage-1);
	    x.dump2Disk();
	    pagesInMem--;
	}
	else
	    pages = null;
    }

    /**
     * This method is used to add a new value to our array. It
     * is assumed that the values are corresponding to
     * consecutive indexes!
     *
     * @param v   The value to be added to the array.
     * @exception AddWhileInGetModeException This means that you tried to add more
     *            values after the object has gone into get mode!
     */
    public void add(subInstance v) throws AddWhileInGetModeException {
	if (getMode) 
	    throw new AddWhileInGetModeException();
	int pId = nextIndex / pageSize;
	if (pId == nextPage) {
	    // Must create a new page!
	    if (pagesInMem < maxPageInMem) {
		Page x = null;
		try {
		    x = new Page(nextPage, pageSize);
		} catch (OutOfMemoryError e) {
		    System.err.println("I am here!");
		    System.exit(100);
		}
		pages.add(x);
		pagesInMem++;
		nextPage++;
		try {
		    x.add(v);
		    nextIndex++;
		} catch (PageNotInMemoryException e) {
		    System.err.println("No way! I just created it!");
		} catch (PageIsFullException e) {
		    System.err.println(e);
		    e.printStackTrace();
		}
	    }
	    else {
		// Write one page to disk and try again!
		freeSpace();
		add(v);
	    }
	}
	else {
	    Page x = (Page)pages.get(pId);
	    try {
		x.add(v);
	    } catch (PageNotInMemoryException e) {
		System.err.println(e);
		e.printStackTrace();
		System.exit(100);
	    } catch (PageIsFullException e) {
		System.err.println(e);
		e.printStackTrace();
	    }
	    nextIndex++;
	}
    }

    /**
     * This method is used to retrieve idx'th value in our array.
     * This value may reside on disk, in which case, it will be retrieved
     * first. Each time a page is needed to be back into memory, the most
     * recently used page will be dumped.
     */
    public subInstance get(int idx) {
	getMode = true;
	int pId = idx / pageSize;
	Page x = (Page)pages.get(pId);
	
	while (true) {
	    try {
		subInstance value = x.get(idx);
		totalAccess++;
		return value;
	    } catch (PageNotInMemoryException e) {
		if (pagesInMem == maxPageInMem) 
		    freeSpace();
		x.back2Mem();
		pagesInMem++;
		pageFaults++;
		sleep = false;
	    }
	}
    }

    /** 
     * This method is called to put the most recently used page to disk, and
     * free enough space for the system to load a new page!
     */
    public void freeSpace() {
	Page x = (Page)Collections.max(pages);
	x.dump2Disk();
	pagesInMem--;
    }


    public void destructor() {
	Iterator it = pages.iterator();
	while (it.hasNext()) {
	    Page x = (Page)it.next();
	    x.destructor();
	}
	pages = null;
    }

    public static void reportFaults() {
	System.err.println("\tTotal disk access: "+totalAccess);
	System.err.println("\tTotal page faults: "+pageFaults);
	if (totalAccess != 0)
	    System.err.println("\tPage faUlt percentage: %"+(int)(100*pageFaults/totalAccess));
    }
    
    private class Page implements Comparable, Serializable {
	private int id;
	private File dump;
	private subInstance[] mem;
	private int used;
	private long lastTimeUsed;

	public Page(int i, int s) {
	    id = i;
	    mem = new subInstance[s];
	    dump = null;
	    used = 0;
	}

	public int compareTo(Object o) {
	    if (!(o instanceof Page))
		throw new ClassCastException();
	    if (lastTimeUsed < ((Page)o).lastTimeUsed) {
		return -1;
	    }
	    else if (lastTimeUsed == ((Page)o).lastTimeUsed) {
		return 0;
	    }
	    else {
		return 1;
	    }
	}


	public void add(subInstance v) throws PageNotInMemoryException, PageIsFullException {
	    if (mem == null)
		throw new PageNotInMemoryException();
	    if (used == mem.length) 
		throw new PageIsFullException("used = "+used);
	    mem[used] = v;
	    used++;
	}

	public subInstance get(int i) throws PageNotInMemoryException {
	    if (mem == null)
		throw new PageNotInMemoryException();
	    lastTimeUsed = System.currentTimeMillis();
	    return mem[i - id*mem.length];
	}

	public boolean dump2Disk() {
	    try {
		dump = newPageFile();
		ObjectOutputStream out = new ObjectOutputStream(new BufferedOutputStream(new FileOutputStream(dump)));
		out.writeObject(mem);
		mem = null;
		out.flush();
		out.close();
		lastTimeUsed = Long.MIN_VALUE;
		return true;
	    } catch (InvalidClassException e) {
		System.err.println(e);
		e.printStackTrace();
	    } catch (NotSerializableException e) {
		System.err.println(e);
		e.printStackTrace();
	    } catch (IOException e) {
		System.err.println(e);
		e.printStackTrace();
	    }
	    return false;
	}

	public boolean back2Mem() {
	    try {
		ObjectInputStream in = new ObjectInputStream(new BufferedInputStream(new FileInputStream(dump)));
		mem = (subInstance [])in.readObject();
		in.close();
		lastTimeUsed = 0;
		dump.delete();
		fileCount--;
		dump = null;
		return true;
	    } catch (ClassNotFoundException e) {
		System.err.println(e);
		e.printStackTrace();	    
	    } catch (InvalidClassException e) {
		System.err.println(e);
		e.printStackTrace();
	    } catch (NotSerializableException e) {
		System.err.println(e);
		e.printStackTrace();
	    } catch (IOException e) {
		System.err.println(e);
		e.printStackTrace();
	    }
	    return false;
	}

	private File newPageFile() throws IOException {
	    File curDir=new File(subGraph.instRoot+"/dump"+Integer.toHexString(fileCount / 255).toUpperCase());
	    if (!curDir.exists()) {
		if (!curDir.mkdirs()) {
		    throw new IOException("Unable to create directory: "+curDir.getPath());
		}
		curDir.deleteOnExit();
	    }
	    fileCount++;
	    return File.createTempFile("dump", ".page", curDir);
	}

	public boolean isInMem() {
	    if (mem == null)
		return false;
	    else
		return true;
	}

	public void destructor() {
	    if (dump != null) {
		dump.delete();
		fileCount--;
	    }
	    mem = null;
	}
    }
}

