import java.io.*;
import java.lang.*;
import java.util.*;

/** 
 * This class implements an array of integers of
 * changing size. The array is partly in main memory
 * and partly saved on disk.
 *
 * @author Shayan Ghazizadeh <shayan@cs.umd.edu>
 */
class IntegerDiskArrayList {

    private ArrayList pages;
    private int maxPageInMem;
    private int pagesInMem;
    private int pageSize;
    private int nextIndex;
    private int nextPage;
    private boolean getMode;

    private int totalAccess;
    private int pageFaults;

    /**
     * Constructor.
     *
     * @param m  Maximum number of pages to have in main memory
     *           at any time.
     * @param ps size of each page.
     */
    public IntegerDiskArrayList(int m, int ps) throws IOException {
	pages = new ArrayList();
	maxPageInMem = m;
	pageSize = ps;
	nextIndex = 0;
	nextPage = 0;
	pagesInMem = 0;

	getMode = false;
	
	totalAccess = pageFaults = 0;
	
	File x = new File(subGraph.instRoot+"/int-dump");
	if (!x.exists()) {
	    if (!x.mkdirs()) {
		    throw new IOException("Unable to create directory: "+x.getPath());
	    }
	    x.deleteOnExit();
	}
	    
    }

    /**
     * This method is used to add a new value to our array. It
     * is assumed that the values are corresponding to
     * consecutive indexes!
     *
     * @param v   The value to be added to the array.
     * @exception AddWhileInGetModeException This means that you tried to add more
     *            values after the object has gone into get mode!
     */
    public void add(int v) throws AddWhileInGetModeException {
	if (getMode) 
	    throw new AddWhileInGetModeException();
	int pId = nextIndex / pageSize;
	if (pId == nextPage) {
	    // Must create a new page!
	    if (pagesInMem < maxPageInMem) {
		Page x = new Page(nextPage, pageSize);
		pages.add(x);
		pagesInMem++;
		nextPage++;
		try {
		    x.add(v);
		    nextIndex++;
		} catch (PageNotInMemoryException e) {
		    System.err.println("No way! I just created it!");
		} catch (PageIsFullException e) {
		    System.err.println(e);
		    e.printStackTrace();
		}
	    }
	    else {
		// Write one page to disk and try again!
		freeSpace();
		add(v);
	    }
	}
	else {
	    Page x = (Page)pages.get(pId);
	    try {
		x.add(v);
	    } catch (PageNotInMemoryException e) {
		System.err.println(e);
		e.printStackTrace();
		System.exit(100);
	    } catch (PageIsFullException e) {
		System.err.println(e);
		e.printStackTrace();
	    }
	    nextIndex++;
	}
    }

    /**
     * This method is used to retrieve idx'th value in our array.
     * This value may reside on disk, in which case, it will be retrieved
     * first. Each time a page is needed to be back into memory, the least
     * recently used page will be dumped.
     */
    public int get(int idx) {
	getMode = true;
	int pId = idx / pageSize;
	Page x = (Page)pages.get(pId);
	
	while (true) {
	    try {
		int value = x.get(idx);
		totalAccess++;
		return value;
	    } catch (PageNotInMemoryException e) {
		freeSpace();
		x.back2Mem();
	    }
	}
    }

    /** 
     * This method is called to put the least recently used page to disk, and
     * free enough space for the system to load a new page!
     */
    public void freeSpace() {
	Page x = (Page)Collections.min(pages);
	x.dump2Disk();
	pagesInMem--;
	pageFaults++;
    }


    /**
     * This method returns the current number of objects in our array.
     */
    public int size() {
	return nextIndex;
    }


    public void destructor() {
	Iterator it = pages.iterator();
	while (it.hasNext()) {
	    Page x = (Page)it.next();
	    x.destructor();
	}
	pages = null;
    }

    public void reportFaults() {
	System.err.println("\tTotal disk access: "+totalAccess);
	System.err.println("\tTotal page faults: "+pageFaults);
	if (totalAccess != 0)
	    System.err.println("\tPage fault percentage: %"+(int)(100*pageFaults/totalAccess));
    }
    
    private class Page implements Comparable {
	private int id;
	private File dump;
	private int[] mem;
	private int used;
	private long lastTimeUsed;

	public Page(int i, int s) {
	    id = i;
	    mem = new int[s];
	    dump = null;
	    used = 0;
	}

	public int compareTo(Object o) {
	    if (!(o instanceof Page))
		throw new ClassCastException();
	    if (lastTimeUsed < ((Page)o).lastTimeUsed) {
		return -1;
	    }
	    else if (lastTimeUsed == ((Page)o).lastTimeUsed) {
		return 0;
	    }
	    else {
		return 1;
	    }
	}


	public void add(int v) throws PageNotInMemoryException, PageIsFullException {
	    if (mem == null)
		throw new PageNotInMemoryException();
	    if (used == mem.length) 
		throw new PageIsFullException("used = "+used);
	    mem[used] = v;
	    used++;
	}

	public int get(int i) throws PageNotInMemoryException {
	    if (mem == null)
		throw new PageNotInMemoryException();
	    lastTimeUsed = System.currentTimeMillis();
	    return mem[i - id*mem.length];
	}

	public boolean dump2Disk() {
	    try {
		dump = newPageFile();
		ObjectOutputStream out = new ObjectOutputStream(new BufferedOutputStream(new FileOutputStream(dump)));
		out.writeObject(mem);
		mem = null;
		out.flush();
		out.close();
		lastTimeUsed = Long.MAX_VALUE;
		return true;
	    } catch (InvalidClassException e) {
		System.err.println(e);
		e.printStackTrace();
	    } catch (NotSerializableException e) {
		System.err.println(e);
		e.printStackTrace();
	    } catch (IOException e) {
		System.err.println(e);
		e.printStackTrace();
	    }
	    return false;
	}

	public boolean back2Mem() {
	    try {
		ObjectInputStream in = new ObjectInputStream(new BufferedInputStream(new FileInputStream(dump)));
		mem = (int [])in.readObject();
		in.close();
		lastTimeUsed = 0;
		dump.delete();
		dump = null;
		return true;
	    } catch (ClassNotFoundException e) {
		System.err.println(e);
		e.printStackTrace();	    
	    } catch (InvalidClassException e) {
		System.err.println(e);
		e.printStackTrace();
	    } catch (NotSerializableException e) {
		System.err.println(e);
		e.printStackTrace();
	    } catch (IOException e) {
		System.err.println(e);
		e.printStackTrace();
	    }
	    return false;
	}

	private File newPageFile() throws IOException {
	    return File.createTempFile("dump", ".page", new File(subGraph.instRoot+"/int-dump"));
	}

	public boolean isInMem() {
	    if (mem == null)
		return false;
	    else
		return true;
	}

	public void destructor() {
	    if (dump != null)
		dump.delete();
	    mem = null;
	}
    }
}

/**
 * This exception is thrown if you try to perform some
 * operation that needs the page to be in memory, but
 * your page is on disk.
 *
 * @author Shayan Ghazizadeh <shayan@cs.umd.edu>
 */
class PageNotInMemoryException extends Exception {
}

/**
 * This exception is thrown if you try to add a new
 * value to a page that is already full.
 *
 * @author Shayan Ghazizadeh <shayan@cs.umd.edu>
 */
class PageIsFullException extends Exception {
    public PageIsFullException() {}
    public PageIsFullException(String msg) {
	super(msg);
    }
}

/**
 * This exception is thrown if you try to perform an add
 * operation while the object is in get mode. When an
 * object goes into get mode, you can not add anymore
 * values to it!
 *
 * @author Shayan Ghazizadeh <shayan@cs.umd.edu>
 */
class AddWhileInGetModeException extends Exception {
}


class SleepWhileInAddModeException extends Exception {
}
