/**
 *
 * SubGraphDiskArrayList.java -- used in SEuS structure discovery system
 * Copyright (C) 2002 Shayan Ghazizadeh
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

import java.io.*;
import java.lang.*;
import java.util.*;

/** 
 * This class implements an array of instances of
 * changing size. The array is partly in main memory
 * and partly saved on disk.
 *
 * @author Shayan Ghazizadeh <shayan@cs.umd.edu>
 */
public class SubGraphDiskArrayList {

    private ArrayList pages;
    private int auxPage;
    private int pageSize;
    private int nextIndex;
    private int nextPage;

    private static int totalAccess=0;
    private static int pageFaults=0;
    private int fileCount;
    /**
     * Constructor.
     *
     * @param m  Maximum number of pages to have in main memory
     *           at any time.
     * @param ps size of each page.
     */
    public SubGraphDiskArrayList(int ps) {
	pages = new ArrayList();
	pageSize = ps;
	nextIndex = 0;
	nextPage = 0;
	auxPage = -1;
	fileCount = 0;
    }

    /**
     * This method returns the current number of objects in our array.
     */
    public int size() {
	return nextIndex;
    }


    /**
     * This method is used to add a new value to our array. It
     * is assumed that the values are corresponding to
     * consecutive indexes!
     *
     * @param v   The value to be added to the array.
     */
    public void add(subGraph v) {
	int pId = nextIndex / pageSize;
	if (pId == nextPage) {
	    // Must create a new page!
	    freeSpace();
	    Page x = null;
	    try {
		x = new Page(nextPage, pageSize);
	    } catch (OutOfMemoryError e) {
		System.err.println("I am here!");
		System.exit(100);
	    }
	    pages.add(x);
	    nextPage++;
	    add(v);
	}
	else {
	    Page x = (Page)pages.get(pId);
	    try {
		x.add(v);
	    } catch (PageNotInMemoryException e) {
		System.err.println(e);
		e.printStackTrace();
		System.exit(100);
	    } catch (PageIsFullException e) {
		System.err.println(e);
		e.printStackTrace();
	    }
	    nextIndex++;
	}
    }

    /**
     * This method is used to retrieve idx'th value in our array.
     * This value may reside on disk, in which case, it will be retrieved
     * first. Each time a page is needed to be back into memory, the most
     * recently used page will be dumped.
     */
    public subGraph get(int idx) {
	int pId = idx / pageSize;
	Page x = (Page)pages.get(pId);
	
	while (true) {
	    try {
		subGraph value = x.get(idx);
		totalAccess++;
		return value;
	    } catch (PageNotInMemoryException e) {
		freeAuxSpace();
		x.back2Mem();
		auxPage = pId;
		pageFaults++;
	    }
	}
    }

    public boolean contains(subGraph s, int start, int end) {

	for (int i=start; i<end; i++) {
	    subGraph s2 = get(i);
	    if (s.equals(s2))
		return true;
	}
	return false;
    }

    /** 
     * This method is called to put the most recently used page to disk, and
     * free enough space for the system to load a new page!
     */
    public void freeSpace() {
	if (nextPage != 0) {
	    Page x = (Page)pages.get(nextPage-1);
	    x.dump2Disk();
	}
    }


    /** 
     * This method is called to put the most recently used page to disk, and
     * free enough space for the system to load a new page!
     */
    public void freeAuxSpace() {
	if (auxPage != -1) {
	    Page x = (Page)pages.get(auxPage);
	    x.dump2Disk();
	}
    }

    public void destructor() {
	Iterator it = pages.iterator();
	while (it.hasNext()) {
	    Page x = (Page)it.next();
	    x.destructor();
	}
	pages = null;
    }

    public static void reportFaults() {
	System.err.println("\tTotal disk access: "+totalAccess);
	System.err.println("\tTotal page faults: "+pageFaults);
	if (totalAccess != 0)
	    System.err.println("\tPage faUlt percentage: %"+(int)(100*pageFaults/totalAccess));
    }
    
    private class Page implements Comparable {
	private int id;
	private File dump;
	private subGraph[] mem;
	private int used;
	private long lastTimeUsed;

	public Page(int i, int s) {
	    id = i;
	    mem = new subGraph[s];
	    dump = null;
	    used = 0;
	}

	public int compareTo(Object o) {
	    if (!(o instanceof Page))
		throw new ClassCastException();
	    if (lastTimeUsed < ((Page)o).lastTimeUsed) {
		return -1;
	    }
	    else if (lastTimeUsed == ((Page)o).lastTimeUsed) {
		return 0;
	    }
	    else {
		return 1;
	    }
	}


	public void add(subGraph v) throws PageNotInMemoryException, PageIsFullException {
	    if (mem == null)
		throw new PageNotInMemoryException();
	    if (used == mem.length) 
		throw new PageIsFullException("used = "+used);
	    mem[used] = v;
	    used++;
	}

	public subGraph get(int i) throws PageNotInMemoryException {
	    if (mem == null)
		throw new PageNotInMemoryException();
	    lastTimeUsed = System.currentTimeMillis();
	    return mem[i - id*mem.length];
	}

	public boolean dump2Disk() {
	    try {
		dump = newPageFile();
		ObjectOutputStream out = new ObjectOutputStream(new BufferedOutputStream(new FileOutputStream(dump)));
		out.writeObject(mem);
		mem = null;
		out.flush();
		out.close();
		lastTimeUsed = Long.MIN_VALUE;
		return true;
	    } catch (InvalidClassException e) {
		System.err.println(e);
		e.printStackTrace();
	    } catch (NotSerializableException e) {
		System.err.println(e);
		e.printStackTrace();
	    } catch (IOException e) {
		System.err.println(e);
		e.printStackTrace();
	    }
	    return false;
	}

	public boolean back2Mem() {
	    try {
		ObjectInputStream in = new ObjectInputStream(new BufferedInputStream(new FileInputStream(dump)));
		mem = (subGraph [])in.readObject();
		in.close();
		lastTimeUsed = 0;
		dump.delete();
		fileCount--;
		dump = null;
		return true;
	    } catch (ClassNotFoundException e) {
		System.err.println(e);
		e.printStackTrace();	    
	    } catch (InvalidClassException e) {
		System.err.println(e);
		e.printStackTrace();
	    } catch (NotSerializableException e) {
		System.err.println(e);
		e.printStackTrace();
	    } catch (IOException e) {
		System.err.println(e);
		e.printStackTrace();
	    }
	    return false;
	}

	private File newPageFile() throws IOException {
	    File curDir=new File(seus.subgraphRoot+"/dump"+Integer.toHexString(fileCount / 255).toUpperCase());
	    if (!curDir.exists()) {
		if (!curDir.mkdirs()) {
		    throw new IOException("Unable to create directory: "+curDir.getPath());
		}
		curDir.deleteOnExit();
	    }
	    fileCount++;
	    return File.createTempFile("dump", ".page", curDir);
	}

	public boolean isInMem() {
	    if (mem == null)
		return false;
	    else
		return true;
	}

	public void destructor() {
	    if (dump != null) {
		dump.delete();
		fileCount--;
	    }
	    mem = null;
	}
    }
}

