/**
 *
 * XML2Graph.java -- converts XML files to be fed into SEuS structure discovery system
 * Copyright (C) 2002 Shayan Ghazizadeh
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This software uses the Sun JavaTM API for XML Processing (JAXP)
 * which is distributed with this software. See the accompanying
 * licences for more details.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

// JAXP packages
import javax.xml.parsers.*;
import org.xml.sax.*;
import org.xml.sax.helpers.*;

// Java packages
import java.util.*;
import java.io.*;


/**
 * This class is used to implement the pending cross-edges,
 * whose real destination is not known yet.
 *
 * @author Shayan Ghazizadeh <shayan@cs.umd.edu>
 */
class pendingEdge {
    // A hash table to store edges. The destination ID is the key
    private Hashtable table;

    /**
     * construtor.
     */
    public pendingEdge() {
	table = new Hashtable();
    }

    /**
     * This method is used to add a new pending edge to
     * the list of edges.
     *
     * @param from  the source of edge (node id in graph file)
     * @param to    the destination of edge (label in XML file, node id is not known yet!)
     * @param label the label for this edge
     */
    public void add(int from, String to, String label) {
	ArrayList x = (ArrayList)table.get(to);
	if (x == null) {
	    x = new ArrayList();
	}
	else {
	    table.remove(to);
	}
	String line = from + " - " + label;
	x.add(line);
	table.put(to, x);
    }

    /**
     * This method is called when the actual id of some node is found.
     * The method will add all of the corresponding pending edges to
     * the edge output file.
     *
     * @param to       the destination of edges (label in XML file)
     * @param actualTo the destination of edges (label in graph file)
     * @param outD     the stream to which the edges are written
     */
    public void resolved(String to, int actualTo, BufferedWriter outD) throws IOException {
	ArrayList x = (ArrayList)table.get(to);
	if (x == null)
	    return;
	// No more unknown!
	table.remove(to);

	// Add pending edges
	Iterator it = x.iterator();
	while (it.hasNext()) {
	    String line = (String)it.next();
	    int idx = line.indexOf('-');
	    String head = line.substring(0, idx-1);
	    String tail = line.substring(idx+1, line.length());
	    outD.write("d "+head+" "+actualTo+tail);
	    outD.newLine();
	}
    }
}


/**
 * This class is used to convert a XML document, to a graph
 * file, which can be fed into the subDis program in order to
 * find frequent subGraphs.
 *
 * @author Shayan Ghazizadeh <shayan@cs.umd.edu>
 */
public class XML2Graph extends DefaultHandler {

    // The output edge definition file
    private BufferedWriter outD;

    // The output vertex definition file
    private BufferedWriter outV;

    // Next available vertex ID
    private int next;

    // Current nodes id!
    private Stack cur;

    // This hashtable is used to store id mappings from
    // XML to graph file!
    private Hashtable idMap;

    // This is a class which saves all the pending edges.
    private pendingEdge pending;

    // Store the wanted vertex output file name.
    private String vFileName;

    /**
     * Constructor.
     *
     * @param outDFile   the edge definition output file name
     * @param outVFile   the vertex definition output file.
     */
    public XML2Graph(String outDFile, String outVFile) {
	next = 1;
	cur = new Stack();
	pending = new pendingEdge();
	try {
	    outD = new BufferedWriter(new FileWriter(outDFile));
	    vFileName = outVFile;
	    outV = new BufferedWriter(new FileWriter(outVFile+".tmp"));
	}
	catch (IOException ioe) {
	    System.err.println(ioe.toString());
	    System.exit(1);
	}
    }


    // Parser calls this once at the beginning of a document
    public void startDocument() throws SAXException {
	idMap = new Hashtable();
    }

    // Parser calls this for each element in a document
    public void startElement(String namespaceURI, String localName,
                             String rawName, Attributes atts) throws SAXException {
	try {
	    if (next != 1) {
		// Add a child edge!
		outD.write("d "+ ((Integer)cur.peek()).intValue() + " " + next + " CHILD");
		outD.newLine();
	    }
	    outV.write("v "+ next + " " + localName);
	    outV.newLine();
	    cur.push(new Integer(next));
	    next++;
	    // Go through atts and treat them as special children!
	    for(int idx=0; idx<atts.getLength(); idx++) {
		String type = atts.getType(idx);
		if (type.equals("ID")) {
		    // Add to mapping
		    String key = atts.getValue(idx);
		    idMap.put(key, (Integer)cur.peek());
		    pending.resolved(key, ((Integer)cur.peek()).intValue(), outD);
		}
		else if (type.equals("IDREF")) {
		    // Create a cross-edge
		    String label = atts.getLocalName(idx).toUpperCase();
		    if (label.equals("IDREF"))
			label = "REF";
		    Integer to = (Integer)idMap.get(atts.getValue(idx));
		    if (to == null) {
			// Make it pending
			pending.add(((Integer)cur.peek()).intValue(), atts.getValue(idx), label); 
		    }
		    else {
			outD.write("d "+((Integer)cur.peek()).intValue()+" "+to.intValue()+" "+label);
			outD.newLine();
		    }
		}
		else if (type.equals("IDREFS")) {
		    // Create cross-edges
		    String label = atts.getLocalName(idx);
		    if (label.toUpperCase().equals("IDREFS"))
			label = "REF";
		    StringTokenizer toks = new StringTokenizer(atts.getValue(idx));
		    String key = null;
		    while (toks.hasMoreTokens()) {
			key = toks.nextToken();
			Integer to = (Integer)idMap.get(key);
			if (to == null) {
			    // Make it pending
			    pending.add(((Integer)cur.peek()).intValue(), key, label); 
			}
			else {
			    outD.write("d "+((Integer)cur.peek()).intValue()+" "+to.intValue()+" "+label);
			    outD.newLine();
			}
		    }
		}
		else {
		    // Treat as normal attribute!
		    outV.write("v "+next+" "+atts.getValue(idx));
		    outV.newLine();
		    outD.write("d "+((Integer)cur.peek()).intValue()+" "+next+" "+atts.getLocalName(idx));
		    outD.newLine();
		    next++;
		}
	    }
	}
	catch (IOException ioe) {
	    System.err.println(ioe.toString());
	    System.exit(1);
	}
    }

    // Parser calls thie for the content of each element!
    public void characters(char[] ch, int start, int length) throws SAXException {
	//	try {
	    String label = (new String(ch, start, length)).trim();
	    if (label.length() > 0) {
		// Change newline to black character!
		label = label.replace('\n', ' ');
		/*
		outV.write("v "+next+" "+label);
		outV.newLine();
		outD.write("d "+((Integer)cur.peek()).intValue()+ " " + next + " VALUE");
		outD.newLine();
		next++;
		*/
	    }
	    /*	}
       	catch (IOException ioe) {
	    System.err.println(ioe.toString());
	    System.exit(1);
	}
	    */
    }

    // Parser calls this after parsing each element in a document 
    public void endElement(String namespaceURI, String localName,
                             String rawName) throws SAXException {
	cur.pop();
    }

    
    // Parser calls this once after parsing a document
    public void endDocument() throws SAXException {
	try {
	    outV.close();
	    outD.close();
	    Process sort = Runtime.getRuntime().exec("sort -k 2,2 -n "+vFileName+".tmp -o "+vFileName);
	    sort.waitFor();
	    (new File(vFileName+".tmp")).delete();
	}
	catch (IOException ioe) {
	    System.err.println(ioe.toString());
	    System.exit(1);
	}
	catch (InterruptedException ie) {
	    System.err.println(ie.toString());
	    System.exit(10);
	}
    }

    /**
     * Convert from a filename to a file URL.
     */
    private static String convertToFileURL(String filename) {
        // On JDK 1.2 and later, simplify this to:
        // "path = file.toURL().toString()".
        String path = new File(filename).getAbsolutePath();
        if (File.separatorChar != '/') {
            path = path.replace(File.separatorChar, '/');
        }
        if (!path.startsWith("/")) {
            path = "/" + path;
        }
        return "file:" + path;
    }

    private static void usage() {
        System.err.println("Usage: XML2GraphV <xml-filename> <output-filename>");
        System.exit(1);
    }


    /**
     * This method is used to create a object of
     * the XML2Graph class type and do all the 
     * necessary initializations, before parsing
     * the XML file. It parses the input file, and
     * generates the output file, in graph format readable
     * by subDis class.
     *
     * @param args an array of string of size two. The first
     *             string is the input XML file name and the second is
     *             the output graph file name.
     */
    static public void convert(String[] args) {
	boolean validation = false;
        // There are several ways to parse a document using SAX and JAXP.
        // We show one approach here.  The first step is to bootstrap a
        // parser.  There are two ways: one is to use only the SAX API, the
        // other is to use the JAXP utility classes in the
        // javax.xml.parsers package.  We use the second approach here
        // because it allows the application to use a platform default
        // implementation without having to specify a system property.
        // After bootstrapping a parser/XMLReader, there are several ways
        // to begin a parse.  In this example, we use the SAX API.

        // Create a JAXP SAXParserFactory and configure it
        SAXParserFactory spf = SAXParserFactory.newInstance();
        spf.setValidating(validation);

        XMLReader xmlReader = null;
        try {
            // Create a JAXP SAXParser
            SAXParser saxParser = spf.newSAXParser();

            // Get the encapsulated SAX XMLReader
            xmlReader = saxParser.getXMLReader();
        } catch (Exception ex) {
            System.err.println(ex);
            System.exit(1);
        }

        // Set the ContentHandler of the XMLReader
        xmlReader.setContentHandler(new XML2Graph(args[1]+".d", args[1]+".v"));

        // Set an ErrorHandler before parsing
        xmlReader.setErrorHandler(new XML2GraphErrorHandler(System.err));

        try {
            // Tell the XMLReader to parse the XML document
            xmlReader.parse(convertToFileURL(args[0]));

	    // Concat the files!
	    BufferedWriter vertex = new BufferedWriter(new FileWriter(args[1]+".v", true));
	    BufferedReader edge = new BufferedReader(new FileReader(args[1]+".d"));
	    int ch;
	    while ((ch = edge.read()) != -1) {
		vertex.write((char)ch+edge.readLine());
		vertex.newLine();
	    }
	    vertex.close();
	    edge.close();
	    (new File(args[1]+".v")).renameTo(new File(args[1]));
	    (new File(args[1]+".d")).delete();
        } catch (SAXException se) {
            System.err.println(se.getMessage());
            System.exit(1);
        } catch (IOException ioe) {
            System.err.println(ioe);
            System.exit(1);
	}
    }

    /**
     * For test only.
     *
     * @param args an string array of size two, with two file names.
     *             the first one is the xml file and the second one is
     *             the output files prefix.
     */
    public static void main(String[] args) {

	if (args.length < 2) {
	    usage();
	}

	convert(args);
    }


    // Error handler to report errors and warnings
    private static class XML2GraphErrorHandler implements ErrorHandler {
        /** Error handler output goes here */
        private PrintStream out;

        XML2GraphErrorHandler(PrintStream out) {
            this.out = out;
        }

        /**
         * Returns a string describing parse exception details
         */
        private String getParseExceptionInfo(SAXParseException spe) {
            String systemId = spe.getSystemId();
            if (systemId == null) {
                systemId = "null";
            }
            String info = "URI=" + systemId +
                " Line=" + spe.getLineNumber() +
                ": " + spe.getMessage();
            return info;
        }

        // The following methods are standard SAX ErrorHandler methods.
        // See SAX documentation for more info.

        public void warning(SAXParseException spe) throws SAXException {
            out.println("Warning: " + getParseExceptionInfo(spe));
        }
        
        public void error(SAXParseException spe) throws SAXException {
            String message = "Error: " + getParseExceptionInfo(spe);
            throw new SAXException(message);
        }

        public void fatalError(SAXParseException spe) throws SAXException {
            String message = "Fatal Error: " + getParseExceptionInfo(spe);
            throw new SAXException(message);
        }
    }
}
