/**
 *
 * dataGuide.java -- summary data structure used in SEuS structure discovery system
 * Copyright (C) 2002 Shayan Ghazizadeh
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

import java.io.*;
import java.util.*;

/**
 * This class creates a dataGuide from 
 * a text file containing a graph definition.
 */
public class dataGuide {
    private static int nodePerDir = 10000;
    private static int nodePerFile = 100;
    public static String indexRoot = null;
    private BufferedReader graphStream;
    private boolean debugMode = false;
    private int sLevel;
    /** 
     * Stores label summaries and vertex mappings.
     */
    private ArrayList labels;
    private ArrayList edges;
    private ArrayList lattices;
    public String filename;

    /**
     * Store the mappings between vertex ids and their labels' index
     * in labels data structure
     */
    private IntegerDiskArrayList vertexID;

    private int lineno;

    public dataGuide(String lfilename, int sl) {
	filename = lfilename;
	try {
	    graphStream = new BufferedReader(new FileReader(filename));
	    labels = new ArrayList();
	    edges = new ArrayList();
	    sLevel = sl;
	    if (sLevel == 1)
		lattices = new ArrayList();
	    else
		lattices = null;
	    vertexID = new IntegerDiskArrayList(10, 10000);
	} catch (FileNotFoundException e) {
	    System.err.println("dataGuide: Graph file <"+filename+"> not found.");
	} catch (IOException e) {
	    System.err.println(e);
	    e.printStackTrace();
	    System.exit(200);
	}
    }

    /** 
     * This method reads the vertex information from the graph file
     * and creates a summary data structure about the vertex labels.
     * The vertex IDs should be sequential!
     */
    public boolean readVertices() {
	int idx;
	int start, end, ch;
	String line;
	int pu=0;
	Integer u;

	System.err.println("dataGuide.readVertices(): adding vertices to data guide.");
	if (graphStream == null)
	    return false;

	try {
	    lineno = 0;
	    /** 
	     * Read the first character line.
	     */
	    while ((ch = graphStream.read()) == 'v') {
		labelNode d = new labelNode();
		/**
		 * Read the vertex id and label.
		 */
		line = graphStream.readLine();
		lineno++;
		start = 0;
		// Skip white space
		while (line.charAt(start) == ' ')
		    start++;

		end = line.indexOf(' ', start);
		u = Integer.valueOf(line.substring(start, end));
		if (u.intValue() != pu+1)
		    throw new IOException();

		start = end;
		while ((start < line.length()) && (line.charAt(start) == ' '))
		    start++;
		end = line.length();
		d.label = line.substring(start, end);
		if (d.label.equals(""))
		    d.label = "<EMPTY>";
		d.occ = 1;
		if ((idx = labels.indexOf(d)) == -1) {
		    // New label
		    labels.add(d);
		    edges.add((ArrayList)null);
		    if (sLevel == 1)
			lattices.add((ArrayList)null);
		    idx = labels.size()-1;
		}
		else {
		    // Label already in list
		    labelNode p = (labelNode)labels.get(idx);
		    p.add(d);
		    labels.set(idx, p);
		}
		vertexID.add(idx);
		pu = u.intValue();
	    }
	    if (ch != 'd')
		throw new IOException();
	    return true;
	} catch (IOException e) {
	    System.err.println("dataGuide.readVertices(): Graph file is corrupt.");
	    if (graphStream != null)
		System.err.println("line:"+lineno);
	    labels = null;
	    return false;
	} catch (AddWhileInGetModeException e) {
	    System.err.println(e);
	    e.printStackTrace();
	    labels = null;
	    return false;
	} catch (Exception e) {
	    System.err.println(e);
	    e.printStackTrace();
	    if (graphStream != null)
		System.err.println("line:"+lineno);
	    labels = null;
	    return false;
	}
    }


    /**
     * This method reads the edges from the graph file and
     * creates the transition matrix for the data guide.
     */
    public boolean readEdges() {
	int u, v;
	int from, to;
	int idx;
	int start, end;
	String line;
	int ch = 'd';

	System.err.println("dataGuide.readEdges(): Adding edges to data guide");
	if (graphStream == null)
	    return false;

	try {
	    //	    System.err.println("# of Labels in dataGuide: "+labels.size());
	    //	    System.err.println("# of vertices in graph  : "+vertexID.size());
	    /** 
	     * Read the first character in a line.
	     */


	    // Set up the index info!
	    File indexDir = new File(indexRoot+"/0");
	    if (!indexDir.exists()) {
		indexDir.mkdirs();
	    }
	    File indexFile = new File(indexDir.getAbsolutePath()+"/0");
	    BufferedWriter indexBuf = new BufferedWriter(new FileWriter(indexFile.getAbsolutePath(), true));
	    
	    int prevU=1;
	    latticeNode latNode = null;
	    if (sLevel == 1) {
		prevU = 1;
		latNode = new latticeNode();
	    }

	    do {
	    
		if ((ch != 'd') && (ch != 'u'))
		    throw new IOException();
		// Read the edge information
		line = graphStream.readLine();
		lineno++;
		start = 0;
		// Skip white space
		while (line.charAt(start) == ' ')
		    start++;
		end = line.indexOf(' ', start);
		u = (new Integer(line.substring(start, end))).intValue();

		start = end;
		// Skip white space
		while (line.charAt(start) == ' ')
		    start++;
		end = line.indexOf(' ', start);
		v = (new Integer(line.substring(start, end))).intValue();

		edgeLabel p = new edgeLabel();
		start = end;
		// Skip white space
		while (line.charAt(start) == ' ')
		    start++;
		end = line.length();
		p.label = line.substring(start, end);
		p.occ = 1;

		// write the index -- FROM SUBGRAPH CLASS!
		if (indexDir.getAbsolutePath().equals(indexRoot+"/"+(u/nodePerDir))) {
		    if (!indexFile.getAbsolutePath().equals(indexDir+"/"+(u/nodePerDir)+"/"+(u/nodePerFile))) {
			indexBuf.close();
			indexFile = new File(indexDir.getAbsolutePath()+"/"+(u/nodePerFile));
			indexBuf = new BufferedWriter(new FileWriter(indexFile.getAbsolutePath(), true));
		    }
		}
		else {
		    indexBuf.close();
		    indexDir = new File(indexRoot+"/"+(u/nodePerDir));
		    if (!indexDir.exists()) {
			indexDir.mkdirs();
		    }
		    indexFile = new File(indexDir.getAbsolutePath()+"/"+(u/nodePerFile));
		    indexBuf = new BufferedWriter(new FileWriter(indexFile.getAbsolutePath(), true));
		}
		indexBuf.write(u + " " + v + " " + p.label);
		indexBuf.newLine();

		// find label indexes for u and v
		from = vertexID.get(u-1);
		to = vertexID.get(v-1);
		p.toIndex = to;
		ArrayList handleFanOut = (ArrayList)edges.get(from);

		if (handleFanOut == null) {
		    // No edges for this connection so far
		    handleFanOut = new ArrayList();
		    handleFanOut.add(p);
		}
		else if ((idx = handleFanOut.indexOf(p)) == -1) {
		    // New label for this connection
		    handleFanOut.add(p);
		}
		else {
		    // Label already exists for this connection
		    edgeLabel q = (edgeLabel)handleFanOut.get(idx);
		    q.add(p);
		    handleFanOut.set(idx, q);
		}
		// Put back the modified label and it's edges ;)
		edges.set(from, handleFanOut);

		// add to the lattice!
		if (sLevel == 1) {
		    if (u == prevU) {
			latNode.add(p.label, getLabel(v-1).label);
		    }
		    else {
			// Add the previous one to the dg
			//System.out.println(u+"\t"+prevU);
			int pFrom = vertexID.get(prevU-1);
			ArrayList latArray = (ArrayList)lattices.get(pFrom);
			if (latArray == null) {
			    latNode.first();
			    latArray = new ArrayList();
			    latArray.add(latNode);
			    lattices.set(pFrom, latArray);
			}
			else {
			    int lidx = latArray.indexOf(latNode);
			    if (lidx == -1) {
				latNode.first();
				latArray.add(latNode);
			    }
			    else {
				latticeNode n = (latticeNode)latArray.get(lidx);
				n.inc();
				latArray.set(lidx, n);
			    }
			}
			// create a new one!
			latNode = new latticeNode();
			latNode.add(p.label, getLabel(v-1).label);
			prevU = u;
		    }
		}

		if (false/*ch == 'u'*/) {
		    // Add the other way too!
		    //		    handle = (labelNode)labels.get(to);
		    handleFanOut = (ArrayList)edges.get(to);
		    p.toIndex = from;

		    if (handleFanOut == null) {
			// No edges for this connection so far
			handleFanOut = new ArrayList();
			handleFanOut.add(p);
		    }
		    else if ((idx = handleFanOut.indexOf(p)) == -1) {
			// New label for this connection
			handleFanOut.add(p);
		    }
		    else {
			// Label already exists for this connection
			edgeLabel q = (edgeLabel)handleFanOut.get(idx);
			q.add(p);
			handleFanOut.set(idx, q);
		    }
		    // Put back the modified label ;)
		    //		    labels.set(to, handle);
		    edges.set(to, handleFanOut);
		}
	    } while ((ch = graphStream.read()) != -1);
	    
	    //Add the last one to the lattice!
	    if (sLevel == 1) {
		int pFrom = vertexID.get(u-1);
		ArrayList latArray = (ArrayList)lattices.get(pFrom);
		if (latArray == null) {
		    latNode.first();
		    latArray = new ArrayList();
		    latArray.add(latNode);
		    lattices.set(pFrom, latArray);
		}
		else {
		    int lidx = latArray.indexOf(latNode);
		    if (lidx == -1) {
			latNode.first();
			latArray.add(latNode);
		    }
		    else {
			latticeNode n = (latticeNode)latArray.get(lidx);
			n.inc();
			latArray.set(lidx, n);
		    }
		}
	    }
	    return true;
	}
	catch (IOException e) {
	    System.err.println("dataGuide.readEdges(): Graph file is corrupt.");
	    if (graphStream != null)
		System.err.println("line:"+lineno);
	    labels = null;
	    edges = null;
	    return false;
	}
    }

    /**
     * This method is passed a list of labelNodes.
     * It will look for all possible edges than can be added to
     * this list of labels according to the current dataGuide configuration.
     * The return value is a list of addableEdge objects.
     */
    public ArrayList findChildren(ArrayList el) {
	ArrayList result = new ArrayList();
	ListIterator it = el.listIterator();
	while (it.hasNext()) {
	    labelNode ln = (labelNode)it.next();
	    int u = labels.indexOf(ln);
	    
	    /*
	    if (u != it.nextIndex()-1)
		continue;
	    */

	    ln = (labelNode)labels.get(u);
	    ArrayList Ledges = (ArrayList)edges.get(u);
	    if (Ledges != null) {
		// There is an edge going out from ln!
		Iterator eit = Ledges.iterator();
		while (eit.hasNext()) {
		    edgeLabel e = (edgeLabel)eit.next();
		    addableEdge newAdd = new addableEdge();
		    newAdd.edge = (edgeLabel)e.clone();
		    newAdd.from = (labelNode)ln.clone();
		    newAdd.to = (labelNode)(((labelNode)labels.get(e.toIndex)).clone());
		    result.add(newAdd);
		}
	    }
	}
	return result;
    }


    public Iterator allVerticeLabels() {
	return labels.iterator();
    }

    /**
     * This method is passed an index in vertexID array, and
     * it returns the label for that id.
     */
    public labelNode getLabel(int u) {
	return (labelNode)labels.get(vertexID.get(u));
    }

    public void clearEdges() {
	edges = null;
    }

    public void cleanUp() {
	vertexID.destructor();
    }

    public void reportDiskAccess(int m) {
	if (m == 0)
	    System.err.println("IntegerArray Disk access summary for creating data guide:");
	else
	    System.err.println("IntegerArray Disk access summary at the end:");
	vertexID.reportFaults();
    }

    public void reportSize() {
	System.err.println("Number of vertices: "+labels.size());
	int c=0;
	ListIterator it = edges.listIterator();
	while (it.hasNext()) {
	    ArrayList l = (ArrayList)it.next();
	    if (l != null)
		c+= l.size();
	}
	System.err.println("Number of edges: "+c);
    }

    public int getHeuristic(int u, subGraph n) {
	latticeNode ln = new latticeNode();
	labelNode from = (labelNode)n.labels.get(u);
 
	Iterator it = ((ArrayList)n.edges.get(u)).iterator();
	while (it.hasNext()) {
	    edgeLabel e = (edgeLabel)it.next();
	    ln.add(e.label, ((labelNode)n.labels.get(e.toIndex)).label);
	}
	
	ArrayList latArray = (ArrayList)lattices.get(labels.indexOf(from));
	int res = 0;
	it = latArray.iterator();
	while (it.hasNext()) {
	    latticeNode node = (latticeNode)it.next();
	    res = res + ln.subset(node)*node.getCount();
	}
	return res;
    }

    public int graphSize() {
	return vertexID.size();
    }
}


