/**
 *
 * seus.java -- Kernel of SEuS structure discovery system
 * Copyright (C) 2002 Shayan Ghazizadeh
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

import java.util.*;
import java.io.*;

/**
 * This class runs a A* search on subgraph search space,
 * using a heuristic available in dataGuide class, to find most
 * frequent subgraph structures.
 */
public class seus {
    public static String subgraphRoot;

    private int m;
    /**
     * A reference to the dataGuide used for this discovery session.
     */
    public dataGuide dg;

    /**
     * List of all subGraphs which are not processed (expanded) yet. 
     */
    private ArrayList open;

    /**
     * List of all subGraphs that are already processed.
     * This is used to prevent the system to search some paths more than once ;)
     */
    private ArrayList close;

    /**
     * This is the set of subGraphs discovered at the end.
     */
    //private ArrayList best;
    private SubGraphDiskArrayList best;
    private Hashtable edgeHash;
    private int maxInBest;
    public ArrayList tempBest;

    /**
     * This is the maximum number of subGraphs that the user is interested
     * in. This should be initialized when generating an object.
     */
    private final int maxBestSize;
    
    /**
     * This is the minimum number of occurances a subGraph should have
     * to be considered a frequent one from users point of view.
     */
    private final int threshold;

    /**
     * This is the Structure Complexity Measure used to control the
     * structure sizes...
     */
    private final double scm;


    private String filename;

    /**
     * Constructor. It will initialize the open list with
     * a subGraph for each available vertex label that
     * has an occurance greater than the threshold. 
     * and sets close and best list to empty.
     */
    public seus(dataGuide ldg, int mbs, int tresh, double lscm, int lbw, String lfilename) {
	maxBestSize = mbs;
	m = mbs * lbw;
	threshold = tresh;
	scm = lscm;
	filename = lfilename;
	dg = ldg;
	open = new ArrayList();
	//open = new TreeSet(new subGraphCompare());
	//close = new ArrayList();
	//best = new ArrayList();
	best = new SubGraphDiskArrayList(500);
	edgeHash = new Hashtable();
	maxInBest = -1;

	Iterator labels = dg.allVerticeLabels();
	while (labels.hasNext()) {
	    labelNode label = (labelNode)labels.next();
	    // <TODO>
	    if (label.occ >= threshold)
		openAdd(new subGraph(label));
	}
    }

    /**
     * This is the main method that does the A* search in
     * the search space!
     */
    private void search(int sLevel) {
	subGraph bestNode;
	ArrayList children;
	subGraphCompare c = new subGraphCompare();
	

	System.err.println("subDis.search(): Search for subGraphs started.");

	/**
	 * For now we will continue until there are more
	 * nodes to expand, and the best list has not reached
	 * the maximum size!
	 */
	while ((open.size() > 0)) {
	    bestNode = (subGraph)open.remove(0);
	    int n = bestNode.nEdges();
	    int start, end;
	    if (n <= maxInBest) {
		start = 0;
		end = best.size();
		if (edgeHash.containsKey(new Integer(n))) {
		    start = ((Integer)edgeHash.get(new Integer(n))).intValue();
		}
		if (edgeHash.containsKey(new Integer(n+1))) {
		    end = ((Integer)edgeHash.get(new Integer(n+1))).intValue();
		}
	    }
	    else {
		start = 0;
		end = 0;
	    }
	    if (!best.contains(bestNode, start, end)) {
		children = bestNode.expand(sLevel, dg);
		if (children == null) {
		    // No possible expansions
		    bestAdd(bestNode);
		    continue;
		}
		Collections.sort(children, c);
		subGraph bestChild = (subGraph)children.get(0);
		if( bestChild.getH()  < threshold) {
		    /**
		     * Any subGraph more specific than the
		     * current one (bestNode) will not satisfy the
		     * condition of deepening, which is based on the
		     * increase in compression by using the new subGraph!
		     * so we recognize bestNode as a candidate for solution.
		     */
		    bestAdd(bestNode);
		}
		else {
		    /**
		     * Add those children to open list that are not already in close or open list
		     * and pass the threshold test.
		     */
		    bestAdd(bestNode);
		    Iterator it = children.iterator();
		    do {
			subGraph nextChild = (subGraph)it.next();
			if (nextChild.getH()  >= threshold) {
			    n = nextChild.nEdges();
			    if (n <= maxInBest) {
				start = 0;
				end = best.size();
				if (edgeHash.containsKey(new Integer(n))) {
				    start = ((Integer)edgeHash.get(new Integer(n))).intValue();
				}
				if (edgeHash.containsKey(new Integer(n+1))) {
				    end = ((Integer)edgeHash.get(new Integer(n+1))).intValue();
				}
			    }
			    else {
				start = 0;
				end = 0;
			    }
			    if (!open.contains(nextChild) && !best.contains(nextChild, start, end))
				openAdd(nextChild);
			}
			else
			    break;
		    } while (it.hasNext());
		}
	    }
	}
    }

    private void searchH(int sLevel) {
	subGraph bestNode;
	ArrayList children;
	subGraphCompare c = new subGraphCompare();
	

	System.err.println("subDis.search(): Search for subGraphs started.");

	/**
	 * For now we will continue until there are more
	 * nodes to expand, and the best list has not reached
	 * the maximum size!
	 */
	int localMaxBestSize = m;
	while ((open.size() > 0) && (best.size() < localMaxBestSize)) {
	    bestNode = (subGraph)Collections.min(open, c);
	    open.remove(bestNode);
	    /*
	    int n = bestNode.nEdges();
	    int start, end;
	    if (n <= maxInBest) {
		start = 0;
		end = best.size();
		if (edgeHash.containsKey(new Integer(n))) {
		    start = ((Integer)edgeHash.get(new Integer(n))).intValue();
		}
		if (edgeHash.containsKey(new Integer(n+1))) {
		    end = ((Integer)edgeHash.get(new Integer(n+1))).intValue();
		}
	    }
	    else {
		start = 0;
		end = 0;
	    }
	    */
	    //if (!best.contains(bestNode, start, end)) {
	    if (!best.contains(bestNode, 0, best.size())) {
		children = bestNode.expand(sLevel, dg);
		if (children == null) {
		    // No possible expansions
		    bestAdd(bestNode);
		    continue;
		}
		Collections.sort(children, c);
		subGraph bestChild = (subGraph)children.get(0);
		if( bestChild.score() * scm < bestNode.score()) {
		    /**
		     * Any subGraph more specific than the
		     * current one (bestNode) will not satisfy the
		     * condition of deepening, which is based on the
		     * increase in compression by using the new subGraph!
		     * so we recognize bestNode as a candidate for solution.
		     */
		    bestAdd(bestNode);
		}
		else {
		    /**
		     * Add those children to open list that are not already in close or open list
		     * and pass the threshold test.
		     */
		    bestAdd(bestNode);
		    Iterator it = children.iterator();
		    do {
			subGraph nextChild = (subGraph)it.next();
			if (nextChild.score() * scm >= bestNode.score()) {
			    /*
			    n = nextChild.nEdges();
			    if (n <= maxInBest) {
				start = 0;
				end = best.size();
				if (edgeHash.containsKey(new Integer(n))) {
				    start = ((Integer)edgeHash.get(new Integer(n))).intValue();
				}
				if (edgeHash.containsKey(new Integer(n+1))) {
				    end = ((Integer)edgeHash.get(new Integer(n+1))).intValue();
				}
			    }
			    else {
				start = 0;
				end = 0;
			    }
			    */
			    //if (!open.contains(nextChild) && !best.contains(nextChild, start, end))
			    if (!open.contains(nextChild) && !best.contains(nextChild, 0, best.size()))
				openAdd(nextChild);
			}
			else
			    break;
		    } while (it.hasNext());
		}
	    }
	}
    }


    private void bestAdd(subGraph s) {
	int idx;
	
	best.add(s);
	int n = s.nEdges();
	maxInBest = n;
	if (!edgeHash.containsKey(new Integer(n))) {
	    edgeHash.put(new Integer(n), new Integer(best.size()-1));
	    //System.err.println("subgraphs with "+n+" edges start from "+(best.size()-1)+".");
	}
	if (best.size() % 10 == 0)
	    System.err.println("subDis.bestAdd(): "+best.size()+"th result found! (Open list size="+open.size()+")\t"+n);
    }


    final private void openAdd(subGraph s) {
	open.add(s);
    }


    /**
     * This method checks to see weather the final best list members
     * are all really frequent subGraphs or not. If they are it
     * returns true, otherwise, it removes the bad nodes from best
     * and returns false. Normally if the return value is false,
     * the caller should call search method again to have the needed
     * best nodes at the end.
     * <TODO>
     */
    private boolean count() {
	int i;
	int up = m;
	
	up = best.size();

	for (i=0; i<up; i++) {
	    if (i % 10 == 0)
		System.err.println("subGraph.countInst(): Counting actual number of instances for "+(i+1)+"th result.");
	    if (((subGraph)best.get(i)).complexity() == 1)
		((subGraph)best.get(i)).hf();
	    else
		((subGraph)best.get(i)).countInst(filename, dg);
	}
	return true;
    }


    /**
     * Prints out the needed command line for running a copy of this class.
     */
    public static void usage() {
	System.err.println("Usage: seus [-o file] [-s file] [-xml] [-nsubs n] [-scm i] [-thresh t] [-l0|-l1] input-file");
	System.err.println("\t -o file  : Output file where the result will be written. stderr is default.");
	System.err.println("\t -s file  : Output file where the statistics will be written. stdout is default");
	System.err.println("\t -xml     : Treat input file as a XML file.");
	System.err.println("\t -nsubs n : Find n sub graphs. default is 1 [valid only in approximate mode]");
	System.err.println("\t -scm i   : The structure complexity measure. default is 1.0 [approximate mode]");
	System.err.println("\t -thresh t: The threshold, percentage of input graph. 1.0 is the default. [Complete mode]");
	System.err.println("\t -l0|-l1  : Select summary level. L0 is default.");
    }


    /**
     * The main method for single iteration! This is the entry place to our system!
     */
    public static void main(String[] argv) {
	int nsubs = 1;
	String filename = null;
	boolean xml = false;
	double lscm = 1;
	int lbw=4;
	double threshP=1.0;
	int thresh=20;
	int sLevel = 0;
	boolean complete = false;

	subGraph.indexRoot = System.getProperty("user.dir")+"/tmp/index";
	subGraph.instRoot = System.getProperty("user.dir")+"/tmp/instances";
	seus.subgraphRoot = System.getProperty("user.dir")+"/tmp/subs";
	dataGuide.indexRoot = System.getProperty("user.dir")+"/tmp/index";
	System.setProperty("java.library.path", System.getProperty("java.library.path")+":"+ System.getProperty("user.dir")+"/");
	// Parse arguments
	for(int i=0; i<argv.length; i++) {
	    if (argv[i].equals("-xml")) {
		// Input file is an XML file, it must be converted first!
		xml = true;
	    }
	    else if (argv[i].equals("-nsubs")) {
		nsubs = Integer.valueOf(argv[i+1]).intValue();
		i++;
	    }
	    else if (argv[i].equals("-thresh")) {
		threshP = Double.valueOf(argv[i+1]).doubleValue();
		complete = true;
		i++;
	    }
	    else if (argv[i].equals("-scm")) {
		lscm = Double.valueOf(argv[i+1]).doubleValue();
		complete = false;
		i++;
	    }
	    else if (argv[i].equals("-bw")) {
		lbw = Integer.valueOf(argv[i+1]).intValue();
		i++;
	    }
	    else if (argv[i].equals("-o")) {
		try {
		    System.setErr(new PrintStream(new FileOutputStream(argv[i+1])));
		    
		} catch (IOException e) {
		    System.err.println("Unable to change error output to "+argv[i+1]+". Ignoring -o");
		}
		i++;
	    }
	    else if (argv[i].equals("-s")) {
		try {
		    System.setOut(new PrintStream(new FileOutputStream(argv[i+1], true)));
		    
		} catch (IOException e) {
		    System.err.println("Unable to change standard output to "+argv[i+1]+". Ignoring -s");
		}
		i++;
	    }
	    else if (argv[i].equals("-l0")) {
		sLevel = 0;
	    }
	    else if (argv[i].equals("-l1")) {
		sLevel = 1;
	    }
	    else
		filename = argv[i];
	}

	if (filename == null) {
	    usage();
	}
	else {
	    long t0 = 0;
	    if (xml) {
		t0 = System.currentTimeMillis();
		String [] args = new String[2];
		args[0] = filename;
		try {
		    args[1] = File.createTempFile("dis", "").toString();
		} catch (IOException io) {
		    System.err.println(io);
		    System.exit(1);
		}
		XML2Graph.convert(args);
		filename = args[1];
	    }
	    long t1 = System.currentTimeMillis();
	    dataGuide dg = new dataGuide(filename, sLevel);
	    if (dg.readVertices() && dg.readEdges()) {
		dg.reportSize();
		dg.reportDiskAccess(0);
		long t2 = System.currentTimeMillis();
		if (complete)
		    thresh = (int)(threshP * dg.graphSize());
		else
		    thresh = 1;
		seus sdm = new seus(dg, nsubs, thresh, lscm, lbw, filename);
		if (complete) {
		    sdm.search(sLevel);
		}
		else {
		    sdm.searchH(sLevel);
		}
		long t3 = System.currentTimeMillis();
		sdm.open = null;
		sdm.close = null;
		sdm.best.freeSpace();
		
		sdm.count();
		long t4 = System.currentTimeMillis();
		dg.reportDiskAccess(1);
		System.err.println("Cleaning up...");
		dg.cleanUp();
		subGraph.removeIndexes();
		subGraph.removeInstances();

		System.err.println("subDis.main(): "+sdm.best.size()+" frequent subGraphs discovered.");
		System.err.println("They are as follows:");
		// TODO
		//if (!complete)
		//  Collections.sort(sdm.best, new subGraphCompare());
		
		int counter = 0;
		int totalSave = 0;
		int totalF = 0;
		int totalH = 0;
		int good = 0;
		if (complete) {
		    while (counter < sdm.best.size()) {
			subGraph sg = sdm.best.get(counter);
			if (sg.getF() >= sdm.threshold) {
			    sg.cout(System.err);
			    totalSave += sg.complexity() * (sg.getF());
			    good++;
			    System.err.println();
			    System.err.println();
			}
			totalF += sg.getF();
			totalH += sg.getH();
			counter++;
		    }
		}
		else {
		    while ((counter < sdm.best.size()) /*&& (counter < sdm.maxBestSize)*/) {
			subGraph sg = sdm.best.get(counter);
			if (sg.getF() != -1) {
			    sg.cout(System.err);
			    totalSave += sg.complexity() * (sg.getF());
			    good++;
			    System.err.println();
			    System.err.println();
			}
			totalF += sg.getF();
			totalH += sg.getH();
			counter++;
		    }
		}
		System.err.println("Dataset:"+filename);
		System.err.println("Score sum: "+totalSave);
		System.err.println("Total number of structures: "+good);
		System.err.println("Total number of structures considered: "+sdm.best.size());
		if (complete)
		    System.err.println("Parameters: threshold = "+sdm.threshold);
		else
		    System.err.println("Parameters: scm = "+sdm.scm+"\tBeam width:"+lbw);
		System.err.println("Total F: "+totalF+"\tTotal H:"+totalH+"\tOverestimation :"+((((double)totalH)-totalF)/totalF));

		System.err.println("Time usage:");
		if (xml) {
		    System.err.println("\tTime used to convert XML to graph     : "+(t1-t0)+" milliseconds.");
		}
		System.err.println("\tTime used to create dataGuide & path Indexes: "+(t2-t1)+" milliseconds.");
		System.err.println("\tTime used to find subGraphs                 : "+(t3-t2)+" milliseconds.");
		System.err.println("\tTime used to count number of instances      : "+(t4-t3)+" milliseconds.");
		System.err.println("\tTotal time used                             : "+(t4-t1)+" milliseconds. ("+((t4-t1)/1000)+" seconds)");

		System.out.println(sdm.scm+"\t"+lbw+"\t"+totalSave+"\t"+(t2-t1)+"\t"+(t3-t2)+"\t"+(t4-t3));

		System.err.println();
		System.err.println("Disk Access summary for counting the instances:");
		InstanceDiskArrayList.reportFaults();
	    }
	    else {
		System.err.println("Unable to create dataGuide. Abnormal exit.");
	    }
	}
    }

    // ================================================ GUI related code =============================================================
    /**
     * Start the process for a paticular file and
     * create the dataGuide!
     * Should be called before the first call to incEngine!
     */
    public static dataGuide init(String filename, int sLevel, boolean xml) {
	subGraph.indexRoot = System.getProperty("user.dir")+"/tmp/index";
	subGraph.instRoot = System.getProperty("user.dir")+"/tmp/instances";
	seus.subgraphRoot = System.getProperty("user.dir")+"/tmp/subs";
	dataGuide.indexRoot = System.getProperty("user.dir")+"/tmp/index";
	System.setProperty("java.library.path", System.getProperty("java.library.path")+":"+ System.getProperty("user.dir")+"/");

	if (xml) {
	    String [] args = new String[2];
	    args[0] = filename;
	    try {
		args[1] = File.createTempFile("dis", "").toString();
	    } catch (IOException io) {
		System.err.println(io);
		System.exit(1);
	    }
	    XML2Graph.convert(args);
	    filename = args[1];
	}
	dataGuide dg = new dataGuide(filename, sLevel);
	if (dg.readVertices() && dg.readEdges()) {
	    	dg.reportSize();
		dg.reportDiskAccess(0);
	}
	else {
	    System.err.println("Unable to create dataGuide. Abnormal exit.");
	    dg = null;
	}
	return dg;
    }

    /**
     * The main method. This is the entry place to our system!
     */

    public static seus doSearch(dataGuide dg, int sLevel, int lbw, double lscm) {
	int nsubs = 4;

	seus sdm = new seus(dg, nsubs, 1, lscm, lbw, dg.filename);
	sdm.searchH(sLevel);
	sdm.open = null;
	sdm.close = null;
	sdm.best.freeSpace();
	
	sdm.tempBest = seus.extractList(sdm);
	return sdm;
    }


    public static ArrayList extractList(seus sdm) {
	ArrayList l = new ArrayList();
	int upto = 4*4;
	if (sdm.best.size() < upto)
	    upto = sdm.best.size();
	for (int i=0; i<upto; i++) {
	    l.add(sdm.best.get(i));
	}
	Collections.sort(l, new subGraphCompare());
	
	return l;
    }

    /**
     * Count throw the interface!
     * Can be called only after a doSearch method call!
     */
    public static ArrayList doCount(seus sdm) {
	sdm.count();
	subGraph.removeInstances();
	
	return sdm.tempBest;
    }


    public static void done(seus sdm) {
	System.err.println("Cleaning up...");
	sdm.dg.cleanUp();
	subGraph.removeIndexes();
    }
}



