/**
 *
 * subGraph.java -- subGraph class implementation used in SEuS structure discovery system
 * Copyright (C) 2002 Shayan Ghazizadeh
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program uses the nauty software (version 2) which is
 * redistributed with this software. This library is copyrighted.
 * See the accompanying copyright and licence for more details.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


import java.util.*;
import java.io.*;


/**
  * Implements the Comparator interface for subGraph class.
  * Objects are compared based on h value. This returns the
  * negative of natural compatision result between numbers
  * to insure descending order in lists.
  */
class subGraphCompare implements Comparator {
    public int compare(Object o1, Object o2) {
	if (!(o1 instanceof subGraph))
	    throw new ClassCastException();
	if (!(o2 instanceof subGraph))
	    throw new ClassCastException();
	if (((subGraph)o1).score() < ((subGraph)o2).score())
	    return 1;
	else if (((subGraph)o1).score() == ((subGraph)o2).score()) 
	    return 0;
	else
	    return -1;
    }
    
    
    public boolean equals(Object o) {
	if (!(o instanceof subGraphCompare))
	    return false;
	return true;
    }
}


/**
  * Implements the Comparator interface for subGraph class.
  * Objects are compared based on the number of edges. This returns the
  * negative of natural compatision result between numbers
  * to insure descending order in lists.
  */
class subGraphCompare2 implements Comparator {
    public int compare(Object o1, Object o2) {
	if (!(o1 instanceof subGraph))
	    throw new ClassCastException();
	if (!(o2 instanceof subGraph))
	    throw new ClassCastException();
	if (((subGraph)o1).nEdges() < ((subGraph)o2).nEdges())
	    return -1;
	else if (((subGraph)o1).score() == ((subGraph)o2).score()) 
	    return 0;
	else
	    return 1;
    }
    
    
    public boolean equals(Object o) {
	if (!(o instanceof subGraphCompare))
	    return false;
	return true;
    }
}


/**
 * This is a general class to store subgraphs.
 * This will be used to store nodes in our search-space.
 */ 
class subGraph implements Serializable {
    private boolean debugMode = false;
    private static int nodePerDir = 10000;
    private static int nodePerFile = 100;
    public static String indexRoot = null;
    public static String instRoot = null;

    /** 
     * Stores label summaries and vertex mappings.
     */
    public ArrayList labels;

    /** 
     * Stores the corresponding fanOut ArrayList for each label!
     */
    public ArrayList edges;


    /**
     * This integer is the heuristic value associated with this
     * subgraph according to the dataGuide.
     * It will be -1 if unknow, and any positive value after being
     * evaluated based on a given dataGuide.
     */
    private int h;
    private int f;
    
    private int[] canon;
    private int[] orbits;
    
    
    private subGraph parentG;
    private InstanceDiskArrayList instances;
	
    /**
     * A get method to return the h value.
     */
    public int getH() {
	return h;
    }
    
    public int getF() {
	return f;
    }
    
    
    public int nEdges() {
	int sum = 0;
	ArrayList x;
	
	for (int i=0; i<labels.size(); i++)
	    if ((x = (ArrayList)edges.get(i)) != null )
		sum += x.size();
	return sum;
    }
    
    /** 
     * Returns a metric of complexity for the subGraph.
     * For now it is the number of vertices and edges.
     * <TODO> <COMPLETE>
     */
    public int complexity() {
	return labels.size() + nEdges();
    }
    
    
    public void hf() {
	f = h;
    }
    
    
    /**
     * the native code to call nauty package driver and get the canonical
     * labeling of the graph!
     * -am: is the adjacency matrix of the graph. the rows are concatenated!
     * -c : is the color partitions. see nauty documentation for more info.
     * -n : number of vertices!
     * -l : number of different vertice labels!
     * -orbits: returns orbits array!
     * -return value: lab
     */
    private native int[] getCanonC(int []am, int[]c, int n, int l, int[] orbits);
    static {
	//System.loadLibrary("getCanonImpl.so");
	System.load(System.getProperty("user.dir")+"/getCanonImpl.so");
    }

    
    
    private void getCanon() {
	ArrayList l = new ArrayList();
	int lc=0;
	int []am = new int[labels.size()*labels.size()];
	int []c  = new int[labels.size()];
	
	
	ListIterator it = labels.listIterator();
	while (it.hasNext()) {
	    String str = ((labelNode)it.next()).label;
	    if (l.indexOf(str) == -1) {
		lc++;
		l.add(str);
	    }
	}
	
	Collections.sort(l);
	
	it = labels.listIterator();
	while (it.hasNext()) {
	    String str = ((labelNode)it.next()).label;
	    c[it.previousIndex()] = l.indexOf(str);
	    
	    ArrayList fanOut = (ArrayList)edges.get(it.previousIndex());
	    if (fanOut != null) {
		ListIterator it2 = fanOut.listIterator();
		while (it2.hasNext()) {
		    int toIndex = ((edgeLabel)it2.next()).toIndex;
		    am[labels.size()*it.previousIndex()+toIndex] = 1;
		}
	    }
	}
	
	int s = labels.size();
	orbits = new int[s];
	canon = getCanonC(am, c, s, lc, orbits);
    }
    
    
    public double score() {
	int tmp = complexity();
	double c;
	
	/*
	if (tmp <= 3)
	    c = 1.0;
	else
	    c = 2.34383 - 0.470683 * tmp + 0.0405402 * tmp * tmp;
	
	if (c < 1.0)
	    c = 1.0;

	*/
	// For now c = 1.0 always!

	if (instances != null) 
	    f = instances.size();

	c = 1.0;
	if (f == -1)
	    return tmp * h / c;
	else
	    return tmp * f;
    }



    /** 
     * Generates a string that can be used as a unique label for
     * this subgraph!
     */
    public String genLabel() {
	String res = new String();
	Iterator it = labels.iterator();
	while(it.hasNext()) {
	    res = res + ((labelNode)it.next()).label + " - ";
	}
	res = res + " -- edges -- ";
	ArrayList x;

	for (int i=0; i<labels.size(); i++) {
	    if ((x = (ArrayList)edges.get(i)) != null ) {
		it = x.iterator();
		res = res + "<";
		while(it.hasNext()) {
		    res = res + ((edgeLabel)it.next()).label + " - ";
		}
		res = res + ">";
	    }
	    else {
		res = res + " - ";
	    }
	}
	return res;
    }


    /**
     * returns true if the subgraph has at least two special vertices in it
     */
    public boolean hasTwo() {
	Iterator it = labels.iterator();
	int count = 0;
	while (it.hasNext()) {
	    if (((labelNode)it.next()).sg != null)
		count++;
	}
	if (count >= 2)
	    return true;
	return false;
    }


    /**
     * This method will return an exact copy of this object.
     * It's an implenetation of Object.clone method.
     */
    public Object clone() {
	subGraph copy = new subGraph();
	copy.h = h;
	copy.f = f;
	copy.labels = (ArrayList)labels.clone();
	copy.edges = new ArrayList();
	copy.canon = canon;
	copy.orbits = orbits;
	copy.parentG = parentG;
	copy.instances = instances;
	Iterator it = edges.iterator();
	while (it.hasNext()) {
	    ArrayList x = (ArrayList)it.next();
	    ArrayList xcopy = null;
	    if (x != null) {
		Iterator it2 = x.iterator();
		xcopy = new ArrayList();
		while (it2.hasNext()) {
		    edgeLabel y = (edgeLabel)it2.next();
		    xcopy.add((edgeLabel)y.clone());
		}
	    }
	    copy.edges.add(xcopy);
	}
	return copy;
    }

    /**
     * This method returns true if the passed object is equal to the 
     * current object. otherwise it returns false. It overrides the
     * Object class's equals method.
     */
    public boolean equals(Object oo) {
	if (!(oo instanceof subGraph))
	    return false;

	subGraph o = (subGraph)oo;
	// Compare number of vertices

	if (o.labels.size() != labels.size())
	    return false;

	if (o.nEdges() != nEdges())
	    return false;

	if (canon == null)
	  getCanon();
	
	if (o.canon == null)
	  o.getCanon();

	for (int i=0; i<labels.size(); i++) {
	    
	    int index = canon[i];
	    int indexO= o.canon[i];
	    
	    if (!((labelNode)labels.get(index)).equals(o.labels.get(indexO)))
		return false;
	    
	    ArrayList fanOutO = (ArrayList)o.edges.get(indexO);
	    ArrayList fanOut = (ArrayList)edges.get(index);
	    
	    if ((fanOut == null) && (fanOutO != null))
		return false;
	    if ((fanOut != null) && (fanOutO == null))
		return false;
	    if (fanOut != null) {
		if (fanOut.size() != fanOutO.size())
		    return false;
		Iterator it = fanOut.iterator();
		while (it.hasNext()) {
		    edgeLabel e = (edgeLabel)it.next();
		    int k=0;
		    for (k=0; k<labels.size(); k++)
			if (canon[k] == e.toIndex) {
			    break;
			}
		    
		    ListIterator nli = fanOutO.listIterator();
		    boolean found = false;
		    while (nli.hasNext() && !found) {
			edgeLabel ee = (edgeLabel)nli.next();
			if (ee.label.equals(e.label)) {
			    if (ee.toIndex == o.canon[k]) {
				found = true;
			    }
			    else {
				int rep = o.orbits[o.canon[k]];
				for (int q = 0; q<labels.size(); q++) {
				    if ((o.orbits[q] == rep) && (ee.toIndex == q)) { //from the same orbit!
					found = true;
					break;
				    }
				}
			    }
			}
		    }
		    if (!found)
			return false;
		}
	    }
	}
	return true;
    }   


    /**
     * This method looks at the members of list (which are subGraphs themselves)
     * to see if this object is a superset of any of them. It will return the index of
     * the subGraphs in list that the object is a superset of.
     *
     * @param list An array list of subGraphs to compare current object with.
     */
    public ArrayList subsets(ArrayList list) {
	ArrayList res = new ArrayList();
	ListIterator li = list.listIterator();
	while (li.hasNext()) {
	    subGraph s = (subGraph)li.next();
	    if (s.isSubset(this))
		res.add(new Integer(li.nextIndex()-1));
	}
	return res;
    }

    /**
     * This method looks at the members of list (which are subGraphs themselves)
     * to see if this object is a subgraph of any of them. It will return 
     * true if a superset is found, otherwise false.
     *
     * @param list An array list of subGraphs to compare current object with.
     */
    public boolean hasSuper(ArrayList list) {
	ListIterator li = list.listIterator();
	while (li.hasNext()) {
	    subGraph s = (subGraph)li.next();
	    if (isSubset(s))
		return true;
	}
	return false;
    }

    /**
     * This method checks to see whether the current object is
     * a subset of parameter s or not. 
     *
     * @param s supergraph candidate to be checked.
     */
    public boolean isSubset(subGraph s) {
	// Compare number of vertices
	if (labels.size() > s.labels.size())
	    return false;
	// Compare vertices and edges
	Iterator it = labels.iterator();
	while(it.hasNext()) {
	    labelNode l = (labelNode)it.next();
	    int idx = s.labels.indexOf(l);
	    if (idx == -1)
		return false;
	    
	    labelNode m = (labelNode)s.labels.get(idx);
	    ArrayList fanOutO = (ArrayList)s.edges.get(idx);
	    idx = labels.indexOf(l);
	    ArrayList fanOut = (ArrayList)edges.get(idx);

	    if ((fanOut != null) && (fanOutO == null))
		return false;
	    if (fanOut != null) {
		if (fanOut.size() > fanOutO.size())
		    return false;
		it = fanOut.iterator();
		while (it.hasNext()) {
		    edgeLabel e = (edgeLabel)it.next();
		    edgeLabel ee = (edgeLabel)e.clone();
		    ee.toIndex = s.labels.indexOf((labelNode)labels.get(e.toIndex));
		    if (!fanOutO.contains(ee))
			return false;
		}
	    }
	}
	return true;
    }

    /**
     * The constructor for this class creates the simplest 
     * type of a subGraph, one with a single vertex and no edges.
     * This vertex has a label which is passed to the constructor.
     */
    public subGraph(labelNode l) {
	labels = new ArrayList();
	edges = new ArrayList();
	edges.add((ArrayList)null);
	labels.add(l);
	h = l.occ;
	canon = null;
	orbits = null;
	parentG = null;
	instances = null;
	f = -1;
    }

    /**
     * This is a dummy internal subGraph constructor.
     */
    private subGraph() {
	h = -1;
	f = -1;
	canon = null;
	orbits = null;
	parentG = null;
	instances = null;
    }

    
    private boolean hasIncoming(int idx, addableEdge ae) {
	ListIterator li1 = labels.listIterator();
	while (li1.hasNext()) {
	    labelNode ln = (labelNode)li1.next();
	    if ((li1.previousIndex() != idx) && (ln.label.equals(((labelNode)labels.get(idx)).label))) {
		ArrayList tmp = (ArrayList)edges.get(li1.previousIndex());
		if (tmp != null) {
		    ListIterator li2 = tmp.listIterator();
		    while (li2.hasNext()) {
			edgeLabel e = (edgeLabel)li2.next();
			if ((e.toIndex == idx) && (e.label.equals(ae.edge.label)))
			    return true;
		    }
		}
	    }
	}
	return false;
    }

    /** 
     * This method expands this subGraph in all possible ways according
     * to the passed dataGuide object. It returns the children in an array list.
     * The returned list will be sorted according to evaluation of children based
     * on the dg.
     */
    public ArrayList expand(int sLevel, dataGuide dg) {
	ArrayList result = new ArrayList();
	ArrayList toDo = dg.findChildren(labels);
	Iterator it = toDo.iterator();

	while(it.hasNext()) {
	    addableEdge ae = (addableEdge)it.next();
	    ListIterator li1 = labels.listIterator();
	    while (li1.hasNext()) {
		labelNode ln1 = (labelNode)li1.next();
		if  (ae.from.equals(ln1)) {
		    // check the chain!
		    if (hasIncoming(li1.previousIndex(), ae))
			continue;
		    // This is a match. Check all of its outgoing edges!
		    ArrayList uFanOut = (ArrayList)edges.get(li1.previousIndex());
		    if (uFanOut == null) {
			// Add it to ln1's edges!
			// To add this edge we have to add a new node!
			subGraph n = (subGraph)clone();
			n.labels.add(ae.to);
			n.edges.add((ArrayList)null);
			n.canon = null;
			n.orbits = null;
			n.parentG = this;
			n.instances = null;
			int s = n.labels.size();
			ae.edge.toIndex = s-1;
			ae.edge.id = nEdges();
			ArrayList newUNodeFanOut = new ArrayList();
			newUNodeFanOut.add(ae.edge);
			n.edges.set(li1.previousIndex(), newUNodeFanOut);
			
			// Set the minimum if occurences as the new value of h for child.
			int min;
			if (sLevel == 1)
			    min = dg.getHeuristic(li1.previousIndex(), n);
			else
			    min = ae.edge.occ;

			if (min > ae.to.occ)
			    min = ae.to.occ;
			if (h > min)
			    n.h = min;
			else
			    n.h = h;
			
			result.add(n);
		    }
		    else {
			ListIterator li2 = uFanOut.listIterator();
			boolean found = false;
			while (li2.hasNext() && !found) {
			    edgeLabel el2 = (edgeLabel)li2.next();
			    labelNode nl2 = (labelNode)labels.get(el2.toIndex);
			    if (el2.label.equals(ae.edge.label) && ae.to.equals(nl2))
				found = true;
			}
			if (!found) {
			    // Add it to ln1's edges!
			    // To add this edge we have to add a new node!
			    subGraph n = (subGraph)clone();
			    n.labels.add(ae.to);
			    n.edges.add((ArrayList)null);
			    n.canon = null;
			    n.orbits = null;
			    n.parentG = this;
			    n.instances = null;
			    int s = n.labels.size();
			    ae.edge.toIndex = s-1;
			    ae.edge.id = nEdges();
			    ArrayList newUNodeFanOut = (ArrayList)n.edges.get(li1.previousIndex());
			    newUNodeFanOut.add(ae.edge);
			    n.edges.set(li1.previousIndex(), newUNodeFanOut);
		    
			    // Set the minimum if occurences as the new value of h for child.
			    int min;
			    if (sLevel == 1)
				min = dg.getHeuristic(li1.previousIndex(), n);
			    else 
				min = ae.edge.occ;

			    if (min > ae.to.occ)
				min = ae.to.occ;
			    if (h > min)
				n.h = min;
			    else
				n.h = h;

			    result.add(n);
			}
		    }
		}
	    }
	}
	if (result.size() == 0)
	    result = null;
	else
	    Collections.sort(result, new subGraphCompare());
	return result;
    }


    public boolean isInside(labelNode n) {
	if (n.sg == null) {
	    Iterator it = labels.iterator();
	    while(it.hasNext()) {
		labelNode ln = (labelNode)it.next();
		if (ln.sg != null) {
		    if (ln.label.indexOf(n.label+" ") != -1)
			return true;
		}
	    }
	    return false;
	}
	Iterator it = labels.iterator();
	while(it.hasNext()) {
	    labelNode ln = (labelNode)it.next();
	    if (n.label.indexOf(ln.label+" ") != -1)
		return true;
	}
	return false;
    }

    /**
     * This method simply outputs a subGraph in a text format.
     * First, all the vertices will be outputed, each on a line similar to the following:
     *    v id label
     * Then the edges will follow in the following format: (edge per line)
     *    d id1 id2 label
     */
    void cout(PrintStream out) {
	ListIterator it = labels.listIterator();
	Iterator it2;
	labelNode l;
	ArrayList lFanOut;
	edgeLabel e;
	int idx;
	while(it.hasNext()) {
	    l = (labelNode)it.next();
	    idx = labels.indexOf(l);
	    out.println("v "+it.nextIndex()+" "+l.label);
	}

	it = edges.listIterator();
	while (it.hasNext()) {
	    lFanOut = (ArrayList)it.next();
	    if (lFanOut != null) {
		it2 = lFanOut.iterator();
		    while(it2.hasNext()) {
			e = (edgeLabel)it2.next();
			out.println("d "+it.nextIndex()+" "+(e.toIndex+1)+" "+e.label);
		    }
	    }
	}
	out.println("h = "+h);
	out.println("f = "+f);
	out.println("Complexity = "+complexity());
	out.println("metric = "+(f == -1?h:f)*complexity());
	//f = -1;
	//out.println("Score = "+score());
    }


    private void createEdgeFile(String graphFile, dataGuide dg, int fromIndex) {
	try {
	    BufferedReader in = new BufferedReader(new FileReader(graphFile));
	    BufferedWriter out = new BufferedWriter(new FileWriter("relative.edge"));

	    // Skip the vertex definitions
	    int ch;
	    int start, end;
	    int u, v;
	    labelNode toLabel, fromLabel;
	    String line;
	    while (((ch = in.read()) != 'd') && (ch != 'u'))
		in.readLine();

	    // Set up the index info!
	    File indexDir = new File(indexRoot+"/0");
	    if (!indexDir.exists()) {
		indexDir.mkdirs();
	    }
	    File indexFile = new File(indexDir.getAbsolutePath()+"/0");
	    BufferedWriter indexBuf = new BufferedWriter(new FileWriter(indexFile.getAbsolutePath(), true));

	    do {
		if ((ch != 'd') && (ch != 'u'))
		    throw new IOException();
		// Read the edge information
		line = in.readLine();
		start = 0;
		// Skip white space
		while (line.charAt(start) == ' ')
		    start++;
		end = line.indexOf(' ', start);
		u = (new Integer(line.substring(start, end))).intValue();

		start = end;
		// Skip white space
		while (line.charAt(start) == ' ')
		    start++;
		end = line.indexOf(' ', start);
		v = (new Integer(line.substring(start, end))).intValue();

		edgeLabel p = new edgeLabel();
		start = end;
		// Skip white space
		while (line.charAt(start) == ' ')
		    start++;
		end = line.length();
		p.label = line.substring(start, end);

		// find labelNodes for u and v
		fromLabel = dg.getLabel(u-1);
		toLabel = dg.getLabel(v-1);
		if (labels.contains(fromLabel) && labels.contains(toLabel)) {
		    int lFrom = labels.indexOf(fromLabel);
		    int lTo = labels.indexOf(toLabel);
		    p.toIndex = lTo;
		    ArrayList xFanOut = (ArrayList)edges.get(lFrom);

		    /**
		     * I do not use contains, because equals for
		     * edgeLabel has different semantics from what I
		     * need here.
		     */
		    if (xFanOut != null) {
			Iterator it = xFanOut.iterator();
			while (it.hasNext()) {
			    edgeLabel e = (edgeLabel)it.next();
			    if (e.label.equals(p.label) && (e.id >= fromIndex)) {
				// Is relative
				line = u + " " + v + " " + p.label;
				out.write(line, 0, line.length());
				out.newLine();
				
				//writeIndex(u, v, p.label); replaced by!
				/*
				if (indexDir.getAbsolutePath().equals(indexRoot+"/"+(u/nodePerDir))) {
				    if (!indexFile.getAbsolutePath().equals(indexDir+"/"+(u/nodePerDir)+"/"+(u/nodePerFile))) {
					indexBuf.close();
					indexFile = new File(indexDir.getAbsolutePath()+"/"+(u/nodePerFile));
					indexBuf = new BufferedWriter(new FileWriter(indexFile.getAbsolutePath(), true));
				    }
				}
				else {
				    indexBuf.close();
				    indexDir = new File(indexRoot+"/"+(u/nodePerDir));
				    if (!indexDir.exists()) {
					indexDir.mkdirs();
				    }
				    indexFile = new File(indexDir.getAbsolutePath()+"/"+(u/nodePerFile));
				    indexBuf = new BufferedWriter(new FileWriter(indexFile.getAbsolutePath(), true));
				}
				indexBuf.write(u + " " + v + " " + p.label);
				indexBuf.newLine();
				*/
				break;
			    }
			}
		    }

		    if (false/*ch == 'u'*/) {
			// Check the otherway too ;)
			lFrom = labels.indexOf(toLabel);
			lTo = labels.indexOf(fromLabel);
			p.toIndex = lFrom;
			xFanOut = (ArrayList)edges.get(lTo);

			/**
			 * I do not use contains, because equals for
			 * edgeLabel has different semantics from what I
			 * need here.
			 */
			if (xFanOut != null) {
			    Iterator it = xFanOut.iterator();
			    while (it.hasNext()) {
				if (((edgeLabel)it.next()).label.equals(p.label)) {
				    // Is relative
				    line = v + " " + u + " " + p.label;
				    out.write(line, 0, line.length());
				    out.newLine();


				    //writeIndex(v, u, p.label); replaced by!
				    /*
				    if (indexDir.getAbsolutePath().equals(indexRoot+"/"+(u/nodePerDir))) {
					if (!indexFile.getAbsolutePath().equals(indexDir+"/"+(u/nodePerDir)+"/"+(u/nodePerFile))) {
					    indexBuf.close();
					    indexFile = new File(indexDir.getAbsolutePath()+"/"+(u/nodePerFile));
					    indexBuf = new BufferedWriter(new FileWriter(indexFile.getAbsolutePath(), true));
					}
				    }
				    else {
					indexBuf.close();
					indexDir = new File(indexRoot+"/"+(u/nodePerDir));
					if (!indexDir.exists()) {
					    indexDir.mkdirs();
					}
					indexFile = new File(indexDir.getAbsolutePath()+"/"+(u/nodePerFile));
					indexBuf = new BufferedWriter(new FileWriter(indexFile.getAbsolutePath(), true));
				    }
				    indexBuf.write(u + " " + v + " " + p.label);
				    indexBuf.newLine();
				    */
				    break;
				}
			    }
			}
		    }
		}
	    } while ((ch = in.read()) != -1);
	    out.close();
	    indexBuf.close();
	}
	catch(IOException e) {
	    System.err.println(e);
	    e.printStackTrace();
	}
    }

    /*
    static public void writeIndex(int u, int v, String l) throws IOException {
	File temp = new File(indexDir+"/"+(u/nodePerDir));
	if (!temp.exists()) {
	    temp.mkdirs();
	}
	
	temp = new File(indexDir+"/"+(v/nodePerDir));
	if (!temp.exists()) {
	    temp.mkdirs();
	}
	
	BufferedWriter f = new BufferedWriter(new FileWriter(indexDir+"/"+(u/nodePerDir)+"/"+(u/nodePerFile)+"to", true));
	f.write(u + " " + v + " " + l);
	f.newLine();
	f.close();

	f = new BufferedWriter(new FileWriter(indexDir+"/"+(v/nodePerDir)+"/to"+(v/nodePerFile), true));
	f.write(u + " " + v + " " + l);
	f.newLine();
	f.close();
    }
    */

    static public void removeIndexes() {
	File[] IndexDirs = (new File(indexRoot)).listFiles();
	for (int i=0; i<IndexDirs.length; i++) {
	    File[] dirFiles = IndexDirs[i].listFiles();
	    for (int j=0; j<dirFiles.length; j++)
		dirFiles[j].delete();
	    IndexDirs[i].delete();
	}
    }

    static public void removeInstances() {
	File[] IndexDirs = (new File(instRoot)).listFiles();
	for (int i=0; i<IndexDirs.length; i++) {
	    File[] dirFiles = IndexDirs[i].listFiles();
	    for (int j=0; j<dirFiles.length; j++)
		dirFiles[j].delete();
	    IndexDirs[i].delete();
	}
    }

    public ArrayList orderEdges() {
	ArrayList preRes = new ArrayList();
	ListIterator li = labels.listIterator();
	while (li.hasNext()) {
	    labelNode x = (labelNode)li.next();
	    ArrayList xFanOut = (ArrayList)edges.get(li.nextIndex()-1);
	    if (xFanOut != null) {
		Iterator it = xFanOut.iterator();
		while(it.hasNext()) {
		    edgeLabel e = (edgeLabel)it.next();
		    e.fromIndex = li.previousIndex();
		    addableEdge xyz = new addableEdge(x, (labelNode)labels.get(e.toIndex), e);
		    preRes.add(xyz);
		}
	    }
	}

	Collections.sort(preRes, new addableEdgeCompare());
	return preRes;
    }

    public void countInst(String graphFile, dataGuide dg) {
	
	ArrayList edges = orderEdges();
	ArrayList toUpdate = new ArrayList();
	int pageSize = 5000;
	int maxInMem = 1;
	int comp = 0;

	// Find out where to start counting from!
	subGraph s = parentG;
	toUpdate.add(this);
	while ((s.complexity() != 1) && (s.instances == null)) {
	    toUpdate.add(0, s);
	    s = s.parentG;
	}

	InstanceDiskArrayList parent = null;
	if (s.complexity() == 1) {
	    parent = new InstanceDiskArrayList(maxInMem, pageSize);
	}
	else {
	    parent = s.instances;
	    comp = s.nEdges();
	}
	InstanceDiskArrayList child = new InstanceDiskArrayList(1, pageSize);
	edgeFile ef = null;

	if (edges.size() == 0) {
	    hf();
	    return;
	}

	// Create relative edge file!
	createEdgeFile(graphFile, dg, comp);

	try {
	    while (comp < edges.size()) {
		addableEdge e = (addableEdge)edges.get(comp);
		//System.err.println("Doing edge "+(comp+1)+"\t Current parent size = "+parent.size());
		comp++;
		
		if (comp == 1) { // First edge!
		    
		    
		    // Go through edge file the first time!
		    ef = new edgeFile();
		    while (ef.nextEdge()) {
			if (e.from.equals(dg.getLabel(ef.u-1)) &&
			    e.to.equals(dg.getLabel(ef.v-1)) &&
			    e.edge.label.equals(ef.l)) {
			   
			    subInstance inst = new subInstance(labels.size());
			    inst.idCount = 2;
			    inst.ids[0] = ef.u;
			    inst.ids[1] = ef.v;
			    child.add(inst);
			}
		    }
		    ef = null;
		}
		else {
		    for (int i=0; i<parent.size(); i++) {
			subInstance a = parent.get(i);
			int v=0;
			if (e.edge.toIndex < a.idCount)
			    v = a.ids[e.edge.toIndex];
			int u = a.ids[e.edge.fromIndex];
			File f = null;

			if (e.edge.toIndex >= a.idCount) { // look for new destination!
			    f = new File(indexRoot+"/"+(u/nodePerDir) + "/" + (u/nodePerFile));
			    if (f.exists()) {
				ef = new edgeFile(f);
				while (ef.nextEdge()) {
				    if ((u == ef.u) &&
					e.from.equals(dg.getLabel(ef.u-1)) &&
					e.to.equals(dg.getLabel(ef.v-1)) &&
					e.edge.label.equals(ef.l)) {
					
					boolean repeated = false;
					for (int rIndex=0; rIndex<a.idCount; rIndex++)
					    if (a.ids[rIndex] == ef.v) {
						repeated = true;
						break;
					    }
					
					if (!repeated) {
					    subInstance childInst = (subInstance)a.clone(1);
					    childInst.ids[e.edge.toIndex] = ef.v;
					    child.add(childInst);
					} 
				    }
				}
			    }
			}
			else { // A new edge for an existing vertex!
			    f = new File(indexRoot+"/"+(u/nodePerDir) + "/" + (u/nodePerFile));
			    if (f.exists()) {
				ef = new edgeFile(f);
				while (ef.nextEdge()) {
				    if ((u == ef.u) && (v == ef.v) &&
					e.from.equals(dg.getLabel(ef.u-1)) &&
					e.to.equals(dg.getLabel(ef.v-1)) &&
					e.edge.label.equals(ef.l)) {
					
					subInstance childInst = (subInstance)a.clone();
					child.add(childInst);
				    }
				}
			    }
			}
		    }
		}
		pageSize = (int)(pageSize * ((double)comp / (comp+1)));
		s = (subGraph)toUpdate.remove(0);
		s.instances = child;
		if ((s.parentG != null) && (s.parentG.instances != null))
		    s.parentG.instances.go2Sleep();

		// Finish the update and exit!
		if (child.size() == 0) {
		    while(toUpdate.size() > 0) {
			s = (subGraph)toUpdate.remove(0);
			s.instances = new InstanceDiskArrayList(true);
		    }
		    parent = child;
		    break;
		}

		parent = child;
		//parent.changeMPM(maxInMem);
		child = new InstanceDiskArrayList(1, pageSize);
	    }
	    f = parent.size();
	    //removeIndexes();
	} catch (OutOfMemoryError e) {
	    System.err.println("Parent size = "+parent.size());
	    System.err.println("Child size = "+child.size());
	    System.err.println("Comp = "+comp);
	    System.err.println("subGraph.countInst(): Could not complete the counting. Skipping this subGraph!");
	    e.printStackTrace();
	    parent.destructor();
	    child.destructor();
	    return;
	} catch (AddWhileInGetModeException e) {
	    System.err.println(e);
	    e.printStackTrace();
	    return;
	} catch (SleepWhileInAddModeException e) {
	    System.err.println(e);
	    e.printStackTrace();
	    return;
	}
    }
}

class instEdge implements Serializable {
    public int u;
    public int v;
    public String l;

    public instEdge(int a, int b, String s) {
	u = a;
	v = b;
	l = s;
    }
}

class edgeFile {
    BufferedReader f;
    public int u;
    public int v;
    public String l;

    public edgeFile() {
	try {
	    f = new BufferedReader(new FileReader("relative.edge"));
	}
	catch (IOException e) {
	    System.err.println("Error trying to read relative edge file!");
	}
    }

    public edgeFile(File ff) {
	try {
	    f = new BufferedReader(new FileReader(ff));
	}
	catch (IOException e) {
	    System.err.println(e);
	    e.printStackTrace();
	}
    }

    public void restart() {
	try {
	    f.close();
	    f = new BufferedReader(new FileReader("relative.edge"));
	}
	catch (IOException e) {
	    System.err.println("Error in resetting file!");
	}
    }


    public boolean nextEdge() {
	String line;
	int start, end;

	try {
	    if ((line = f.readLine()) != null) {
		// Read the edge information
		start = 0;
		// Skip white space
		while (line.charAt(start) == ' ')
		    start++;
		end = line.indexOf(' ', start);
		u = (new Integer(line.substring(start, end))).intValue();

		start = end;
		// Skip white space
		while (line.charAt(start) == ' ')
		    start++;
		end = line.indexOf(' ', start);
		v = (new Integer(line.substring(start, end))).intValue();

		start = end;
		// Skip white space
		while (line.charAt(start) == ' ')
		    start++;
		end = line.length();
		l = line.substring(start, end);
		return true;
	    }
	    return false;
	} catch (IOException e) {
	    System.err.println("Error reading next edge!");
	} catch (OutOfMemoryError e) {
	    System.err.println("It is here!");
	}
	return false;
    }
}
